import Cocoa

let bs = [1, 2, 3].map { String($0) }
bs

let cs = [1,2,3,4,5].filter { $0 > 3 }
cs

/// Simple function wrapper, so we can write methods on it.
struct Pred<A> {
    let pred: (A) -> Bool
}

extension Array {
    /// Same filter as normal, just taking a Pred wrapper.
    func filter_(pred: Pred<Element>) -> Array<Element> {
        return self.filter { (element) -> Bool in
            return pred.pred(element)
        }
    }
}

extension Pred {
    func map<B>(_ transform: @escaping (A) -> B) -> Pred<B> {
        return Pred<B> { b -> Bool in
            // What do we have at hand?
            //   b
            //   a -> b
            //   a -> Bool
            // There is no way to combine those to return a `Bool`.
            // `map` is not possible.
            fatalError()
        }
    }

    func pullback<B>(_ transform: @escaping (B) -> A) -> Pred<B> {
        return Pred<B> { b -> Bool in
            let a = transform(b)
            let bool = self.pred(a)
            return bool
        }
    }
}

// Like line 6, but with a Pred.
let ds = [1,2,3,4,5].filter_(pred: Pred<Int>(pred: { $0 > 3 }))
ds

let intsGT3 = Pred<Int>(pred: { $0 > 3 })
let es = [1,2,3,4,5].filter_(pred: intsGT3)

let stringsGT3 = Pred<String>(pred: { str -> Bool in
    guard let int = Int(str) else { return false }
    return int > 3
})

// We want to go from Pred<Int> -> Pred<String>
// so we pull back from String -> Int
let stringsGT3_: Pred<String> = intsGT3.pullback { str -> Int in
    return Int(str) ?? 0
}

let fs = ["1","2","3","4","5"]
    .filter_(pred: stringsGT3_)

let gs = ["1","2","3","4","5","izza", "kevin", "goran"]
    .filter_(pred: stringsGT3_)

// Now pull back from Double -> Int to get a Pred<Double>
let dubsGT3: Pred<Double> = intsGT3.pullback { (double) -> Int in
    return Int(double)
}

let hs = [1.0,2.2,3.14159,4.5,555555].filter_(pred: dubsGT3)

// Question? Why is `3.14159` not in `hs`?

/// Final note: like `map` is to Functor like `pullback` is to Contravariant Functor.
/// Instead of `pullback` you may also see `contramap`; they are synonyms.
/// https://en.wikipedia.org/wiki/Functor
